import pandas as pd
import requests
import base64
import os
import urllib3
import io
urllib3.disable_warnings()
import datetime
import json

class VisierDataConnector:
    """Class to manage ingestion of Visier data connectors

    See Visier user documentation for more information (you will need to log in to the service portal first): 
    https://my.visier.com/csm?id=kb_article_view&sysparm_article=KB0014851

    Parameters
    ----------
    company : str
        The company name used to connect to the Visier application e.g. [company].visier.com
    
    api_key : str
        The API key generated in the admin panel of the Visier application

    user : str
        The username used to connect to the data_connectors (typically email). This user must 
        have access to all data connectors that you intend to connect to with this Class

    pword : str
        The associated password for :code:`user`

    proxies : dict
        Dictionary containing 'http_proxy' and 'https_proxy' keys
    
    sslVerify: bool
        Defaults to True. If set to False, bypasses HTTPS Certificate verification (very dangerous)
    """
    def __init__(self, company, api_key, user, pword, proxies=None, sslVerify=True):

        self.company = company
        self.api_key = api_key
        self.proxies = proxies
        self.sslVerify = sslVerify

        self.auth = "Basic {}==".format(base64.b64encode(bytes("{}:{}".format(user,pword),"utf-8")).decode("utf-8"))

        self.url_base = f"https://{company}.api.visier.io/api/dataconnector/getData"

    def get_connector(self, connector_id, fName=None):
        """Function to make API call to Visier Data Connectors

        Parameters
        ----------
        connector_id: String
            String of id of data connector, available from Visier

        fName: String, optional
            Full file path, including csv extension that you want to give your file if you want to save it. If not declared, no file will be saved.

        Returns
        -------
        df: Pandas Dataframe
            Pandas Dataframe of returned data from the connector

        """
        urllib3.disable_warnings()
        params = {
                'id': connector_id,
                'apikey': self.api_key,
                'a':'b'
            }

        header = {"Authorization":self.auth}

        save=True
        if fName is None:
            save=False
            fName='temp.csv'

        r = requests.get(
            self.url_base, 
            params=params, 
            headers=header, 
            verify=self.sslVerify,
            proxies=self.proxies)

        if r.status_code == 200:
            f=open(fName,'wb')
            f.write(r.content)
            f.close()
        else:
            raise ValueError(f'API call failed with status code: {r.status_code}')

        df = pd.read_csv(fName)

        if not save:
            os.remove(fName)
        
        return df
    

class VisierAPI:
    """
        Class to help manage the authentication side of dealing with 
        the Visier Data Query API endpoints

        Parameters
        ----------
        uname: str
            The username of the API user

        pword: str
            The password of the API user

        apikey: str
            The company-wide API key accessed via the Visier Studio
            admin panel. Contact your admin if you are not sure.

        vanity_name: str
            The name associated with a Visier Cloud tenant. For example, 
            a tenant for an organization called Jupiter would have the 
            vanity name Jupiter. All tenants should use their own vanity 
            name in the API URL. When making changes to an analytic tenant 
            that you manage, the tenant is specified by its tenant code 
            in the API call.

            E.g. https://<<vanity_name>>.visier.com
        """
    def __init__(self, uname, pword, apikey, vanity_name='experian', proxies=None):
        
        self.uname = uname
        self.pword = pword
        self.apikey = apikey
        self.baseurl = f'https://{vanity_name}.api.visier.io/v1'
        self.proxies = proxies
        self._get_security_code()
    

    def _get_security_code(self):
        data = {
            'username': self.uname,
            'password': self.pword
        }

        resp = requests.post(
            f'{self.baseurl}/admin/visierSecureToken',
            data=data,
            proxies=self.proxies
        )

        if resp.status_code==200:
            self.security_code = resp.content.decode('ascii')
            self.security_code_expires = datetime.datetime.now() + datetime.timedelta(minutes=4, seconds=45)

        else:
            raise Exception(
                f'Security Code failed to generate with html status code: {resp.status_code}; Request URL: {self.baseurl}'
            )

    def query_api(self, api_path, method='get', data=None, custom_headers=None):
        """Function to query Visier Data Query APIs

        Parameters
        ----------
        api_path: Str
            The path of the url you wish to query. Must begin with '/'
            e.g. '/data/model/analytic-objects/Employee/dimensions'

        method: Str must be either 'get' or 'post'
            The type of http request you are trying to make

        data: Dictionary
            If your 'method' is 'post', this is the data you wish send 
            as part of the request.
        """
        headers = {
            'apikey': self.apikey,
            'Content-type': 'application/json',

        }

        if custom_headers is not None:
            for k,v in custom_headers.items():
                headers[k] = v

        cookies = {}

        if datetime.datetime.now() < self.security_code_expires:
            cookies['VisierASIDToken'] =  self.security_code
        else:
            self._get_security_code()
            cookies['VisierASIDToken'] =  self.security_code


        if method=='get':
            resp = requests.get(
                f'{self.baseurl}{api_path}',
                headers=headers,
                cookies=cookies,
                proxies=self.proxies
                )

        elif method=='post':
            resp = requests.post(
                f'{self.baseurl}{api_path}',
                headers=headers,
                cookies=cookies,
                data=data,
                proxies=self.proxies)

        if resp.ok:
            return resp
        else:
            raise Exception(f"""
                API call failed with html status code: {resp.status_code}\n
                {resp.json()}
            """)

    def _ttm_split(self, string):
        try:
            return string.split('/')[1]
        except:
            return string

    def _generate_df_from_agg_query(self, resp):
        data = io.StringIO(resp.content.decode('utf-8').replace('\r',''))

        df = pd.read_csv(data, dtype='str')

        df['DateInRange'] = df['DateInRange'].apply(lambda x: x.split(' - ')[0])
        df['DateInRange'] = df['DateInRange'].apply(lambda x: self._ttm_split(x))
        df['DateInRange'] = pd.to_datetime(df['DateInRange'])
        df['DateInRange'] = df['DateInRange'] - datetime.timedelta(days=1)
        df['DateInRange'] = df['DateInRange'].dt.strftime('%Y-%m-%d')

        df = df.fillna('')
        cols = list(df.columns)

        df = df.groupby(cols[cols.index('DateInRange'):]).sum().reset_index()

        return df 

    def get_aggregation_query(self, query):

        resp = self.query_api(
            '/data/query/aggregate',
            method='post',
            data=json.dumps(query),
            custom_headers={
                'Accept':'text/csv'
            }
        )

        df = self._generate_df_from_agg_query(resp)

        return df
    
    def get_list_query(self, query):
        resp = self.query_api(
            '/data/query/list',
            method='post',
            data=json.dumps(query),
            custom_headers={
                'Accept':'text/csv'
            }
        )

        data = io.StringIO(resp.content.decode('utf-8').replace('\r',''))

        df = pd.read_csv(data, dtype='str')

        return df
    
    def get_properties_list(self, analyticObjectId):
        resp = self.query_api(
            f'/data/model/analytic-objects/{analyticObjectId}/properties',
            method='get',
        )
        
        try:
            return resp.json()
        except:
            return resp.content


import pysftp
import os
from datetime import datetime as datetime
import logging

class SFTP:
    """Helper class to manage connections to an SFTP site, file collection and upload

    Parameters
    ----------
    host : str
        The host address you wish to use as the SFTP connection.

    user : str
        The username of the SFTP account you are connecting to.

    pword : str
        The password of the SFTP account you are connecting to.

    port : int
        The port through which the connection to the SFTP server is made.
        522 is used as the default port.
    """

    # Datetime is used to append to folder names
    TodayDate = datetime.now()

    date_time = TodayDate.strftime('%m-%d-%y_%H-%M-%S')

    def __init__(self, host, user, pword, port, date_time=date_time):

        self.host = host
        self.user = user
        self.pword = pword
        self.port = port
        self.date_time = date_time

    def download_file(self, 
                     master_folder, 
                     sub_folder, 
                     local_file=False, 
                     remote_file=False, 
                     return_file_path=False, 
                     return_folder_path=False, 
                     delete_original=True, 
                     use_date_stamp=True):
        """Collects file from STS account and saves within folder location

        Package can also currently used to create folders without interaction with SFTP site
        ----------
        master_folder: String ()
            Path to the root directory for the process
        sub_folder: String ()
            Name used for the sub folder within the root directory. This is where the output will be saved.
            Example: Inputting "Output Files" will create a separate folder "Output Files" within the 
            root directory for files to be saved.
        local_file: String () default = False
            Path to where the file will be saved locally.
        remote_file: String () default = False
            Path to where the file is stored in the SFTP portal.
        return_file_path: Boolean default = False
            Determines whether the filepath to the outputs is returned by the function or not.
        return_folder_path: Boolean default = False
            Determines whether the folder path to the outputs is returned by the function or not.
        delete_original: Boolean default = False
            Determines whether the file is removed from the SFTP portal after it has been downloaded.
        use_date_stamp: Boolean default = True
            Determines whether a date stamp is appended to the sub folder name or not.
        Returns
        -------
        Dependant on the parameters, the function will optionally return the filepath of output 
        and/or path to the sub folder where output is stored.
        """

        if use_date_stamp and sub_folder:
            folder_location = "{0}_{1}".format(sub_folder, self.date_time)
        elif sub_folder:
            folder_location = "{0}".format(sub_folder)
        else:
            folder_location = ""
        folder = r'{0}\\{1}'.format(master_folder, folder_location)

        if not os.path.exists(folder):
            logging.debug(f"Creating folder directory at {folder}")
            os.makedirs(folder)
            logging.debug(f"Successfully created folder directory at {folder}")

        if local_file:
            # Define the local path where the file will be saved
            localFilePath = r'{0}\\{1}'.format(folder, local_file)

        if remote_file and local_file:
            self.__file_collect(remote_file, localFilePath)

            if delete_original:
                self.__file_remove(remote_file)

        if return_file_path and return_folder_path and local_file:
            return localFilePath, folder
        elif return_file_path and local_file:
            return localFilePath
        elif return_folder_path:
            return folder
        else:
            return None

    def upload_file(self, local_file_path, remote_file_path):
        """Uploads file to SFTP account

        ----------
        local_file_path: String ()
            Path to where the file is saved locally.
        remote_file_path: String ()
            Path to where the file will be stored in the SFTP portal.

        Returns
        -------
        Nothing is returned by the function.

        """
        logging.debug(f"Uploading local file: {local_file_path}")
        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None

        logging.debug(f"Establishing new SFTP connection with {self.host}")

        with pysftp.Connection(
            host=self.host,
            username=self.user,
            password=self.pword,
            cnopts=cnopts,
            port=self.port
        ) as sftp:
            sftp.put(local_file_path, remote_file_path, confirm=False)

        logging.debug(f"Successfully uploaded {local_file_path}")

    def list_files(self, remote_dir="/"):
        """Lists file directory within SFTP account

        ----------
        remote_dir: String () default = '/'
            This is the folder name that you would like to see a directory for,
            change to see more/less granular directories.

        Returns
        -------
        Function returns a list of files within the directory.

        """
        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None

        with pysftp.Connection(
            host=self.host,
            username=self.user,
            password=self.pword,
            cnopts=cnopts,
            port=self.port
        ) as sftp:

            files = sftp.listdir(remote_dir)

        return files

    # Code to collect file from SFTP site based and save at local path
    def __file_collect(self, remote_file_path, local_file_path):
        """Function to collect files from SFTP account

        ----------
        remote_file_path: String ()
            Path to where the file is stored in the SFTP portal.
        local_file_path: String ()
            Path to where the file will be saved locally.

        Returns
        -------
        Nothing is returned by the function.

        """

        logging.debug(f"Downloading remote file: {remote_file_path}")

        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None

        logging.debug(f"Establishing new SFTP connection with {self.host}")

        with pysftp.Connection(
            host=self.host,
            username=self.user,
            password=self.pword,
            cnopts=cnopts,
            port=self.port
        ) as sftp:

            sftp.get(remote_file_path, local_file_path)

        logging.debug(f"Successfully downloaded {remote_file_path}")

    # Code to remove file from SFTP site once it has been downloaded
    def __file_remove(self, remote_file_path):
        """Function to remove files from SFTP account

        ----------
        remote_file_path: String ()
            Path to where the file is stored in the SFTP portal.
        Returns
        -------
        Nothing is returned by the function.

        """

        logging.debug(f"Removing remote file: {remote_file_path}")

        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None

        logging.debug(f"Establishing new SFTP connection with {self.host}")

        with pysftp.Connection(
            host=self.host,
            username=self.user,
            password=self.pword,
            cnopts=cnopts,
            port=self.port
        ) as sftp:

            sftp.remove(remote_file_path)

        logging.debug(f"Successfully removed {remote_file_path}")



import pandas as pd
import pkg_resources
from subprocess import call
from datetime import datetime

def csv_to_utf8(file):
    """Converts a csv file to utf-8 encoding

    Parameters
    ----------
    file: String
        Path to file to be encoded 

    Returns
    -------
    fName:
        String of path to reformatted file
    """
    df = pd.read_csv(
        file,
        encoding = 'latin'
    )

    fName = file[:-4] + '-reformatted.csv'

    df.to_csv(
        fName,
        index=False
    )

    print(f'{fName} created')

    return fName

def getDateTimeStamp(time=True):
    """Function to generate a Date/Time stamp for the current time

    Parameters
    ----------
    time: Boolean
        Set to True to include time in Date stamp. False returns just date. Defaults to True.

    Returns
    -------
    stamp:
        String of date-time stamp
    """
    now = datetime.now()
    stamp = str(now.year) + str(now.month).zfill(2) + str(now.day).zfill(2)
    if time:
        stamp = stamp + "_" + str(now.hour).zfill(2) + str(now.minute).zfill(2)
    return stamp





import urllib3
urllib3.disable_warnings()
from fake_useragent import UserAgent

def get_random_useragent(browser=None):
    '''Function to generate a random user agent for web scraping

    Parameters
    ----------
    browser: String
        Valid values are: 'ie', 'opera', 'chrome', 'firefox', 'safari'

    Returns
    ----------
    String of valid User Agent

    '''

    ua = UserAgent()

    useragent = ua.random

    if browser is not None:
        useragent = ua[browser]

    return useragent





from datetime import datetime as datetime
import logging
import os
import sys

class Log_File:
    """Python class to enable logging of key events to a file of your choosing, 
    as well as the console if you wish

    Parameters
    ----------
    directory : str
        The file location where you want the log file saved, 
        excluding the file name

    fName : str (optional)
        The filename, including extension that you want to use as your output.
        If this isn't specified, the file will be called 'progress_log - YYYY-MM-DD HH:MM:SS.txt', 
        where the date & time stamp will be the time when the log is first initiated.
    
    log_level : str (optional)
        Defaults to 'INFO'. Valid options are NOTSET, DEBUG, INFO, WARN, ERROR, CRITICAL
    
    log_fmt : str (optional)
        See https://docs.python.org/3/library/logging.html#logging.LogRecord for definitions 
        of attributes. Defaults to '%(asctime)s: %(levelname)s %(message)s' e.g. 
        '2024-03-14 10:56:54,329: INFO test info message'

    print_to_console : bool (optional)
        Defaults to True. If set to true, will print messages to both the log file, 
        and the console. If False, will only output to the log file    
    """


    def __init__(self, 
                 directory, 
                 fName=None, 
                 log_level = 'INFO', 
                 log_fmt='%(asctime)s: %(levelname)s %(message)s',
                 print_to_console=False):

        self.run_date_time = datetime.now().strftime('%Y-%m-%d T%H-%M-%S')
        self.fName = f'progress_log-{self.run_date_time}.log'

        if fName is not None:
            self.fName = fName
        self.directory = directory
        self.fPath = r'{0}/{1}'.format(directory, self.fName)
        print(self.fPath)

        self.log_level = logging.getLevelName(log_level)
        self.log_fmt = log_fmt
        self.print_to_console = print_to_console

        if not os.path.exists(self.directory):
            os.makedirs(self.directory)
        
        # if not os.path.exists(self.fPath):
        #     with open(self.fPath, 'w'): pass

        self.__create_logger()

    def __create_logger(self):

        #Set Up Logger and level
        log = logging.getLogger()
        log.setLevel(self.log_level)

        #Set log message format
        format=logging.Formatter(self.log_fmt)

        #Set Up log file handler
        file_handler = logging.FileHandler(self.fPath)
        file_handler.setLevel(self.log_level)
        file_handler.setFormatter(format)
        log.addHandler(file_handler)

        #Set up Print to console if set in Class
        if self.print_to_console:
            console_handler = logging.StreamHandler(sys.stdout)
            console_handler.setLevel(self.log_level)
            console_handler.setFormatter(format)
            log.addHandler(console_handler)
            
        self.logger = log
    
    def debug(self, msg):
        self.logger.debug(msg)
    def info(self, msg):
        self.logger.info(msg)
    def warning(self, msg):
        self.logger.warning(msg)
    def error(self, msg):
        self.logger.error(msg)
    def critical(self, msg):
        self.logger.critical(msg)